@@ -1,6 +1,5 @@
 #include "config.h"
 
-#include <libgnome/libgnome.h>
 #include <vicious.h>
 #include "gdm.h"
 #include "misc.h"
@@ -19,15 +18,42 @@
 extern int GdmFirstVT;
 extern gboolean GdmVTAllocation;
 
-#ifdef __linux__
+#if defined (__linux__) || defined (__FreeBSD__) || defined(__DragonFly__)
 
+#ifdef __linux__
 #include <sys/vt.h>
+#endif
+#if defined(__FreeBSD__) || defined(__DragonFly__)
+#include <sys/consio.h>
+#endif
+
+#if defined( __FreeBSD__) || defined(__DragonFly__)
+static const char*
+__itovty(int val)
+{
+	static char str[8];
+	char* next = str + sizeof (str) - 1;
+
+	*next = '\0';
+	do {
+		*--next = "0123456789abcdefghigklmnopqrstuv"[val % 32];
+	} while (val /= 32);
+
+	return next;
+}
+#endif
 
 static int
 open_vt (int vtno)
 {
-	char *vtname = g_strdup_printf ("/dev/tty%d", vtno);
+	char *vtname;
 	int fd;
+
+#if defined (__linux__)
+	vtname = g_strdup_printf ("/dev/tty%d", vtno);
+#elif defined (__FreeBSD__) || defined(__DragonFly__)
+	vtname = g_strdup_printf ("/dev/ttyv%s", __itovty(vtno - 1));
+#endif
 	do {
 		errno = 0;
 		fd = open (vtname, O_RDWR
@@ -45,7 +71,8 @@ get_free_vt (int *vtfd)
 {
 	int fd, fdv;
 	int vtno;
-	GList *to_close_vts = NULL, *li;
+	unsigned short vtmask;
+	struct vt_stat vtstat;
 
 	*vtfd = -1;
 
@@ -60,7 +87,14 @@ get_free_vt (int *vtfd)
 	if (fd < 0)
 		return -1;
 
-	if ((ioctl(fd, VT_OPENQRY, &vtno) < 0) || (vtno == -1)) {
+	if (ioctl (fd, VT_GETSTATE, &vtstat) < 0) {
+		VE_IGNORE_EINTR (close (fd));
+		return -1;
+	}
+
+	for (vtno = GdmFirstVT, vtmask = 1 << (vtno-1);
+			vtstat.v_state & vtmask; vtno++, vtmask <<= 1);
+	if (!vtmask) {
 		VE_IGNORE_EINTR (close (fd));
 		return -1;
 	}
@@ -70,35 +104,7 @@ get_free_vt (int *vtfd)
 		VE_IGNORE_EINTR (close (fd));
 		return -1;
 	}
-
-	while (vtno < GdmFirstVT) {
-		int oldvt = vtno;
-		to_close_vts = g_list_prepend (to_close_vts,
-					       GINT_TO_POINTER (fdv));
-
-		if ((ioctl(fd, VT_OPENQRY, &vtno) < 0) || (vtno == -1)) {
-			vtno = -1;
-			goto cleanup;
-		}
-
-		if (oldvt == vtno) {
-			vtno = -1;
-			goto cleanup;
-		}
-
-		fdv = open_vt (vtno);
-		if (fdv < 0) {
-			vtno = -1;
-			goto cleanup;
-		}
-	}
-
 	*vtfd = fdv;
-
-cleanup:
-	for (li = to_close_vts; li != NULL; li = li->next) {
-		VE_IGNORE_EINTR (close (GPOINTER_TO_INT (li->data)));
-	}
 	return vtno;
 }
 
@@ -145,7 +151,11 @@ gdm_change_vt (int vt)
 int
 gdm_get_cur_vt (void)
 {
+#if defined (__linux__)
 	struct vt_stat s;
+#elif defined (__FreeBSD__) || defined(__DragonFly__)
+	int vtno;
+#endif
 	int fd;
 
 	do {
@@ -158,15 +168,32 @@ gdm_get_cur_vt (void)
 	} while G_UNLIKELY (errno == EINTR);
 	if (fd < 0)
 		return -1;
-
+#if defined (__linux__)
 	ioctl (fd, VT_GETSTATE, &s);
 
 	VE_IGNORE_EINTR (close (fd));
 
 	/* debug */
+	/*
 	printf ("current_Active %d\n", (int)s.v_active);
+	*/
 
 	return s.v_active;
+#elif defined (__FreeBSD__) || defined(__DragonFly__)
+	if (ioctl (fd, VT_GETACTIVE, &vtno) == -1) {
+		VE_IGNORE_EINTR (close (fd));
+		return -1;
+	}
+
+	VE_IGNORE_EINTR (close (fd));
+
+	/* debug */
+	/*
+	printf ("current_Active %d\n", vtno);
+	*/
+
+	return vtno;
+#endif
 }
 
 #else /* here this is just a stub, we don't know how to do this outside
